import java.util.concurrent.locks.ReentrantLock;

/**
 * @author tcrain
 * 
 * @param <K>
 * @param <V>
 */

public class CFTree implements IntSet {

	public static final int NOT_REMOVED = 0;
	public static final int BY_LEFT_ROT = 1;
	public static final int REMOVED = 2;

	static final boolean useFairLocks = false;
	// we encode directions as characters
	static final char Left = 'L';
	static final char Right = 'R';

	private class MaintenanceThread extends Thread {
		CFTree set;

		MaintenanceThread(CFTree set) {
			this.set = set;
		}

		public void run() {
			set.doMaintenance();
		}
	}

	private class MaintVariables {
		long propogations = 0, rotations = 0;
	}

	private final MaintVariables vars = new MaintVariables();

	static class Node {
		int key;

		class BalanceVars {
			volatile int localh, lefth, righth;
		}

		final BalanceVars bal = new BalanceVars();
		volatile Node left;
		volatile Node right;
		final ReentrantLock lock;
		volatile int removed;
		volatile boolean deleted;

		Node(final int key) {
			this.key = key;
			this.removed = NOT_REMOVED;
			this.deleted = false;
			this.lock = new ReentrantLock(useFairLocks);
			this.right = null;
			this.left = null;
			this.bal.localh = 1;
			this.bal.righth = 0;
			this.bal.lefth = 0;
		}

		Node(final int key, final int localh, final int lefth,
				final int righth, final Node left, final Node right,
				final int removed, final boolean deleted) {
			this.key = key;
			this.bal.localh = localh;
			this.bal.righth = righth;
			this.bal.lefth = lefth;
			this.left = left;
			this.right = right;
			this.lock = new ReentrantLock(useFairLocks);
			this.removed = removed;
			this.deleted = deleted;
		}

		void updateLocalh() {
			this.bal.localh = Math.max(this.bal.lefth + 1, this.bal.righth + 1);
		}

	}

	// state
	private final Node root = new Node(Integer.MIN_VALUE);
	volatile boolean stop = false;
	private MaintenanceThread mainThd;
	// used in the getSize function
	int size;

	// Constructors
	public CFTree() {
		// temporary
		this.startMaintenance();
	}

	Node getNext(Node n, int k) {
		int rem = n.removed;
		if (rem == BY_LEFT_ROT)
			return n.right;
		else if (rem == REMOVED)
			return n.left;
		else if (n.key > k)
			return n.right;
		else if (n.key < k)
			return n.left;
		else
			// n.key == k
			return null;
	}

	boolean validate(Node n, int k) {
		Node next = null;
		if (n.removed != NOT_REMOVED)
			return false;
		else if (n.key == k)
			// if (n.key == k)
			return true;
		else if (n.key > k)
			next = n.right;
		else
			next = n.left;

		if (next == null)
			return true;
		return false;
	}

	@Override
	public boolean contains(int key) {
		// Node next, current;
		// next = root;
		// // final Comparable<? super K> k = comparable(key);
		// int rightCmp;
		//
		// // int nodesTraversed = 0;
		//
		// while (true) {
		// current = next;
		// // if (current.key == null) {
		// // rightCmp = -100;
		// // } else {
		// // rightCmp = k.compareTo(current.key);
		// // }
		//
		// if(current.key > key)
		// rightCmp = 100;
		// else if(current.key < key)
		// rightCmp = -100;
		// else
		// rightCmp = 0;
		//
		// if (rightCmp == 0) {
		// // V value = current.value;
		// if (current.deleted) {
		// // if (TRAVERSAL_COUNT) {
		// // finishCount(nodesTraversed);
		// // }
		// return false;
		// }
		// // if (TRAVERSAL_COUNT) {
		// // finishCount(nodesTraversed);
		// // }
		// return true;
		// }
		// if (rightCmp <= 0) {
		// next = current.left;
		// } else {
		// next = current.right;
		// }
		// // if (TRAVERSAL_COUNT) {
		// // nodesTraversed++;
		// // }
		// if (next == null) {
		// // if (TRAVERSAL_COUNT) {
		// // finishCount(nodesTraversed);
		// // }
		// return false;
		// }
		// }

		Node next, node;
		node = root;

		while (true) {
			next = getNext(node, key);
			if (next == null)
				break;
			node = next;
		}
		if (node.key == key && !node.deleted)
			return true;
		return false;

	}

	@Override
	public boolean remove(int key) {
		Node next, node;
		node = root;

		while (true) {

			next = getNext(node, key);
			if (next == null) {
				node.lock.lock();
				if (validate(node, key))
					break;
				node.lock.unlock();
			} else
				node = next;
		}

		boolean result = false;

		if (node.key == key) {
			if (!node.deleted) {
				node.deleted = true;
				result = true;
			}
		}
		node.lock.unlock();
		return result;
		
		// Node next, current;
		// next = root;
		// // final Comparable<? super K> k = comparable(key);
		// int rightCmp;
		// // V value;
		//
		// while (true) {
		// current = next;
		// // if (current == null) {
		// // rightCmp = -100;
		// // } else {
		// // rightCmp = k.compareTo(current.key);
		// // }
		//
		// if(current.key > key)
		// rightCmp = 100;
		// else if(current.key < key)
		// rightCmp = -100;
		// else
		// rightCmp = 0;
		//
		// if (rightCmp == 0) {
		// if (current.deleted) {
		// return false;
		// }
		// current.lock.lock();
		// if (current.removed == NOT_REMOVED) {
		// break;
		// } else {
		// current.lock.unlock();
		// }
		// }
		// if (rightCmp <= 0) {
		// next = current.left;
		// } else {
		// next = current.right;
		// }
		// if (next == null) {
		// if (rightCmp != 0) {
		// return false;
		// }
		// // this only happens if node is removed, so you take the
		// // opposite path
		// // this should never be null
		// System.out.println("Going right");
		// next = current.right;
		// }
		// }
		// // value = current.value;
		// if (current.deleted) {
		// current.lock.unlock();
		// return false;
		// } else {
		// current.deleted = true;
		// current.lock.unlock();
		// // System.out.println("delete");
		// return true;
		// }
	}

	@Override
	public boolean add(int key) {
		Node next, node;
		node = root;

		while (true) {
			next = getNext(node, key);
			if (next == null) {
				node.lock.lock();
				if (validate(node, key))
					break;
				node.lock.unlock();
			} else
				node = next;
		}

		boolean result = false;

		if (node.key == key) {
			if (node.deleted) {
				node.deleted = false;
				result = true;
			}
		} else {
			Node n = new Node(key);
			if (node.key > key)
				node.right = n;
			else
				node.left = n;
			result = true;
		}
		node.lock.unlock();
		return result;
		// int rightCmp;
		// Node next, current;
		// next = root;
		// // final Comparable<? super K> k = comparable(key);
		// Node n = null;
		// // int traversed = 0;
		// // V val;
		//
		// while (true) {
		// current = next;
		// // traversed++;
		// // if (current.key == null) {
		// // rightCmp = -100;
		// // } else {
		// // rightCmp = k.compareTo(current.key);
		// // }
		//
		// if(current.key > key)
		// rightCmp = 100;
		// else if(current.key < key)
		// rightCmp = -100;
		// else
		// rightCmp = 0;
		//
		// if (rightCmp == 0) {
		// // val = current.value;
		// if (!current.deleted) {
		// // System.out.println(traversed);
		// return true;
		// }
		// current.lock.lock();
		// if (current.removed == NOT_REMOVED) {
		// break;
		// } else {
		// current.lock.unlock();
		// }
		// }
		// if (rightCmp <= 0) {
		// next = current.left;
		// } else {
		// next = current.right;
		// }
		// if (next == null) {
		// if (n == null && false) {
		// n = new Node(key);
		// }
		// current.lock.lock();
		// if (current.removed == NOT_REMOVED) {
		// if (rightCmp <= 0) {
		// next = current.left;
		// } else {
		// next = current.right;
		// }
		// if (next == null) {
		// break;
		// } else {
		// current.lock.unlock();
		// }
		// } else {
		// current.lock.unlock();
		// // maybe have to check if the other one is still null before
		// // going the opposite way?
		// // YES!! We do this!
		// if (rightCmp <= 0) {
		// next = current.left;
		// } else {
		// next = current.right;
		// }
		// if (next == null) {
		// if (rightCmp > 0) {
		// next = current.left;
		// } else {
		// next = current.right;
		// }
		// }
		// }
		// }
		// }
		// // val = current.value;
		// if (rightCmp == 0) {
		// if (current.deleted) {
		// current.deleted = false;
		// current.key = key;
		// current.lock.unlock();
		// // System.out.println("insert");
		// // System.out.println(traversed);
		// return true;
		// } else {
		// current.lock.unlock();
		// return false;
		// }
		// } else {
		// if (true) {
		// n = new Node(key);
		// }
		// if (rightCmp <= 0) {
		// current.left = n;
		// } else {
		// current.right = n;
		// }
		// current.lock.unlock();
		// // System.out.println(traversed);
		// // System.out.println("insert");
		// return true;
		// }
	}

	// maintenance
	boolean removeNode(Node parent, char direction) {
		Node n, child;
		// can get before locks because only maintenance removes nodes
		if (parent.removed != NOT_REMOVED)
			return false;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return false;
		// get the locks
		n.lock.lock();
		parent.lock.lock();
		if (!n.deleted) {
			n.lock.unlock();
			parent.lock.unlock();
			return false;
		}
		if ((child = n.left) != null) {
			if (n.right != null) {
				n.lock.unlock();
				parent.lock.unlock();
				return false;
			}
		} else {
			child = n.right;
		}
		if (direction == Left) {
			parent.left = child;
		} else {
			parent.right = child;
		}
		n.left = parent;
		n.right = parent;
		n.removed = REMOVED;
		n.lock.unlock();
		parent.lock.unlock();
		// System.out.println("removed a node");
		// need to update balance values here
		if (direction == Left) {
			parent.bal.lefth = n.bal.localh - 1;
		} else {
			parent.bal.righth = n.bal.localh - 1;
		}
		parent.updateLocalh();
		return true;
	}

	int rightRotate(Node parent, char direction, boolean doRotate) {
		Node n, l, lr, r, newNode;
		if (parent.removed != NOT_REMOVED)
			return 0;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return 0;
		l = n.left;
		if (l == null)
			return 0;
		if (l.bal.lefth - l.bal.righth < 0 && !doRotate) {
			// should do a double rotate
			return 2;
		}
		parent.lock.lock();
		n.lock.lock();
		l.lock.lock();
		lr = l.right;
		r = n.right;
		newNode = new Node(n.key, Math.max(1 + l.bal.righth, 1 + n.bal.righth),
				l.bal.righth, n.bal.righth, lr, r, n.removed, n.deleted);
		l.right = newNode;

		n.removed = REMOVED;
		if (direction == Left) {
			parent.left = l;
		} else {
			parent.right = l;
		}
		l.lock.unlock();
		n.lock.unlock();
		parent.lock.unlock();

		// need to update balance values
		l.bal.righth = newNode.bal.localh;
		l.updateLocalh();
		if (direction == Left) {
			parent.bal.lefth = l.bal.localh;
		} else {
			parent.bal.righth = l.bal.localh;
		}
		parent.updateLocalh();
		// System.out.println("right rotate");
		return 1;
	}

	int leftRotate(Node parent, char direction, boolean doRotate) {

		Node n, r, rl, l, newNode;
		if (parent.removed != NOT_REMOVED)
			return 0;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return 0;
		r = n.right;
		if (r == null)
			return 0;
		if (r.bal.lefth - r.bal.righth > 0 && !doRotate) {
			// should do a double rotate
			return 3;
		}
		parent.lock.lock();
		n.lock.lock();
		r.lock.lock();
		rl = r.left;
		l = n.left;

		newNode = new Node(n.key, Math.max(1 + r.bal.lefth, 1 + n.bal.lefth),
				n.bal.lefth, r.bal.lefth, l, rl, n.removed, n.deleted);

		r.left = newNode;

		// // temp (Need to fix this!!!!!!!!!!!!!!!!!!!!)
//		n.right = parent;
//		n.left = parent;

		n.removed = BY_LEFT_ROT;

		if (direction == Left) {
			parent.left = r;
		} else {
			parent.right = r;
		}
		r.lock.unlock();
		n.lock.unlock();
		parent.lock.unlock();
		// need to update balance values
		r.bal.righth = newNode.bal.localh;
		r.updateLocalh();
		if (direction == Left) {
			parent.bal.lefth = r.bal.localh;
		} else {
			parent.bal.righth = r.bal.localh;
		}
		parent.updateLocalh();
		// System.out.println("left rotate");
		return 1;
	}

	boolean propagate(Node node) {
		Node lchild, rchild;

		lchild = node.left;
		rchild = node.right;

		if (lchild == null) {
			node.bal.lefth = 0;
		} else {
			node.bal.lefth = lchild.bal.localh;
		}
		if (rchild == null) {
			node.bal.righth = 0;
		} else {
			node.bal.righth = rchild.bal.localh;
		}

		node.updateLocalh();

		if (Math.abs(node.bal.righth - node.bal.lefth) >= 2)
			return true;
		return false;
	}

	boolean performRotation(Node parent, char direction) {
		int ret;
		Node node;

		ret = singleRotation(parent, direction, false, false);
		if (ret == 2) {
			// Do a LRR
			node = direction == Left ? parent.left : parent.right;
			ret = singleRotation(node, Left, true, false);
			if (ret > 0) {
				if (singleRotation(parent, direction, false, true) > 0) {
					// System.out.println("LRR");
				}
			}
		} else if (ret == 3) {
			// Do a RLR
			node = direction == Left ? parent.left : parent.right;
			ret = singleRotation(node, Right, false, true);
			if (ret > 0) {
				if (singleRotation(parent, direction, true, false) > 0) {
					// System.out.println("RLR");
				}
			}
		}
		if (ret > 0)
			return true;
		return false;
	}

	int singleRotation(Node parent, char direction, boolean leftRotation,
			boolean rightRotation) {
		int bal, ret = 0;
		Node node, child;

		node = direction == Left ? parent.left : parent.right;
		bal = node.bal.lefth - node.bal.righth;
		if (bal >= 2 || rightRotation) {
			// check reiable and rotate
			child = node.left;
			if (child != null) {
				if (node.bal.lefth == child.bal.localh) {
					ret = rightRotate(parent, direction, rightRotation);
				}
			}
		} else if (bal <= -2 || leftRotation) {
			// check reliable and rotate
			child = node.right;
			if (child != null) {
				if (node.bal.righth == child.bal.localh) {
					ret = leftRotate(parent, direction, leftRotation);
				}
			}
		}
		return ret;
	}

	boolean recursivePropagate(Node parent, Node node, char direction) {

		Node left, right;

		if (node == null)
			return true;
		left = node.left;
		right = node.right;

		if (node.removed == NOT_REMOVED && node.deleted
				&& (left == null || right == null) && node != this.root) {
			if (removeNode(parent, direction)) {
				return true;
			}
		}

		if (stop) {
			return true;
		}

		if (node.removed == NOT_REMOVED) {
			if (left != null) {
				recursivePropagate(node, left, Left);
			}
			if (right != null) {
				recursivePropagate(node, right, Right);
			}
		}

		if (stop) {
			return true;
		}

		// no rotations for now
		if (node.removed == NOT_REMOVED && node != this.root) {
			if (propagate(node)) {
				this.performRotation(parent, direction);
			}
		}

		return true;
	}

	public boolean stopMaintenance() {
		this.stop = true;
		try {
			this.mainThd.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return true;
	}

	public boolean startMaintenance() {
		this.stop = false;

		mainThd = new MaintenanceThread(this);

		mainThd.start();

		return true;
	}

	boolean doMaintenance() {
		long depth = 0, num_iterations = 0;
		while (!stop) {
			recursivePropagate(this.root, this.root.left, Left);
//			for(int i = 0; i<1000000; i++)
//				delayCount++;
			depth += recursiveDepth(this.root);
			num_iterations++;
		}
		System.out.println("NUmber of iterations: " + num_iterations + "average depth: " + depth/num_iterations + "\n");
		return true;
	}

	// not thread safe
	public int getSize() {
		this.size = 0;
		recursiveGetSize(root.left);
		return size;
	}

	void recursiveGetSize(Node node) {
		if (node == null)
			return;
		if (node.removed != NOT_REMOVED) {
			// System.out.println("Shouldn't find removed nodes in the get size function");
		}
		if (!node.deleted) {
			this.size++;
		}
		recursiveGetSize(node.left);
		recursiveGetSize(node.right);
	}

	public int getBalance() {
		int lefth = 0, righth = 0;
		if (root.left == null)
			return 0;
		lefth = recursiveDepth(root.left.left);
		righth = recursiveDepth(root.left.right);
		return lefth - righth;
	}

	int recursiveDepth(Node node) {
		if (node == null) {
			return 0;
		}
		int lefth, righth;
		lefth = recursiveDepth(node.left);
		righth = recursiveDepth(node.right);
		return Math.max(lefth, righth) + 1;
	}

	public void clear() {
		this.stopMaintenance();
		this.resetTree();
		this.startMaintenance();

		return;
	}

	private void resetTree() {
		this.vars.propogations = 0;
		this.vars.rotations = 0;
		root.left = null;
	}

	@Override
	public void begin() {

	}

	@Override
	public void commit() {
		// TODO Auto-generated method stub

	}

	@Override
	public void abort() {
		// TODO Auto-generated method stub

	}

	@Override
	public boolean nontransactionalAdd(int item) {
		// TODO Auto-generated method stub
		return add(item);
	}

	public void recprint(Node p) {
		if (p == null)
			return;
		recprint(p.right);
		System.out.print(p.key + "  ");
		recprint(p.left);
	}

	public int recgetvalue(Node p) {
		if (p == null)
			return 0;
		if (p.deleted || p == root)
			return recgetvalue(p.right) + recgetvalue(p.left);
		else
			return recgetvalue(p.right) + p.key + recgetvalue(p.left);
	}

	public void recursiveAdd(int start, int end) {
		if (start > end)
			return;

		int mid = (start + end) / 2;
		nontransactionalAdd(mid * 2);
		recursiveAdd(start, mid - 1);
		recursiveAdd(mid + 1, end);

		return;
	}

	@Override
	public int finish() {
		// TODO Auto-generated method stub
		stopMaintenance();
		return recgetvalue(root);
	}

	@Override
	public void start(int range) {
		// TODO Auto-generated method stub
		recursiveAdd(0, range / 2 - 1);
	}
}
