//package org.deuce.benchmark.intset;

import java.util.Random;

public class SetBenchmark implements Benchmark {
	IntSet m_set;
	int m_range = 1 << 16;
	int m_rate = 20;
	int m_ops = 1;
	int nb_threads = 1;
	final static int MAX_OPERATIONS_PER_TRANSACTION = 30;
	
	int final_set_value, calculated_set_value;

	public void init(String[] args, int nb_threads) {
		boolean error = false;
		int initial = 256;
		this.nb_threads = nb_threads;
		if (args.length > 0) {
			if (args[0].equals("LockFreeList"))
				m_set = new LockFreeList();
			else if (args[0].equals("LazyList"))
				m_set = new LazyList();
			else if (args[0].equals("PTBList"))
				m_set = new PTBLinkedList();
			else if (args[0].equals("OTBList"))
				m_set = new OTBLinkedList();
			else if (args[0].equals("LazySkipList"))
				m_set = new LazySkipList();
			else if (args[0].equals("PTBSkipList"))
				m_set = new PTBSkipList();
			else if (args[0].equals("OTBSkipList"))
				m_set = new OTBSkipList();
			else if (args[0].equals("CFTree"))
				m_set = new CFTree();
			else if (args[0].equals("PTBTree"))
				m_set = new PTBTree();
			else if (args[0].equals("OTBTree"))
				m_set = new OTBTree();
			else if (args[0].equals("OOTBTree"))
				m_set = new OTBTreeOptimized();
			else if (args[0].equals("BOOTBTree"))
				m_set = new OTBTreeOptimizedBackoff();
			else
				error = true;
		} else
			error = true;
		for (int i = 1; i < args.length && !error; i++) {
			if (args[i].equals("-r")) {
				if (++i < args.length)
					m_range = Integer.parseInt(args[i]);
				else
					error = true;
			} else if (args[i].equals("-w")) {
				if (++i < args.length)
					m_rate = Integer.parseInt(args[i]);
				else
					error = true;
			} else if (args[i].equals("-o")) {
				if (++i < args.length)
					m_ops = Integer.parseInt(args[i]);
				else
					error = true;
			} else
				error = true;
		}
		if (error) {
			System.out
					.println("Benchmark arguments: (LazyLinkedList|...) [-o ops-per-transaction] [-i initial-size] [-r range] [-w write-rate]");
			System.exit(1);
		}

//		// Tree is different
//		if(m_set instanceof OTBTree || m_set instanceof CFTree)
//			m_set.nontransactionalAdd(m_range);
//		else
//			for (int i = 0; i < m_range; i+=2)
//				m_set.nontransactionalAdd(i);
		
		m_set.start(m_range);
		calculated_set_value = ((m_range/2) * (m_range/2 - 1));
		
		System.out.println("Range               = " + m_range);
		System.out.println("Write rate          = " + m_rate + "%");
		System.out.println("Ops per transaction	= " + m_ops);
		System.out.println();
	}

	public BenchmarkThread createThread(int i, int nb) {
		return new SetThread(i, m_set, m_range, m_rate, m_ops, nb);
	}

	public String getStats(BenchmarkThread[] threads) {
		int add = 0;
		int remove = 0;
		int succ_add = 0;
		int succ_remove = 0;
		int succ_contains = 0;
		int contains = 0;
		int aborts = 0;
		int falseinvals = 0;
		
		for (int i = 0; i < threads.length; i++) {
			add += ((SetThread) threads[i]).m_nb_add;
			remove += ((SetThread) threads[i]).m_nb_remove;
			contains += ((SetThread) threads[i]).m_nb_contains;
			succ_add += ((SetThread) threads[i]).m_nb_succ_add;
			succ_remove += ((SetThread) threads[i]).m_nb_succ_remove;
			succ_contains += ((SetThread) threads[i]).m_nb_succ_contains;
			aborts += ((SetThread) threads[i]).m_nb_aborts;
			falseinvals += ((SetThread) threads[i]).m_nb_falseinvals;
			calculated_set_value += ((SetThread) threads[i]).m_thread_contribution;
		}
		
		if(final_set_value != calculated_set_value)
			System.out.println("\n\nERRORR. final = " + final_set_value + " , calculated = " + calculated_set_value + " , difference = " + (final_set_value - calculated_set_value) + "\n\n");
		else
			System.out.println("\n\nPASSED\n\n");
		
		return "A=" + add + ", R=" + remove + ", SA=" + succ_add + ", SR="
				+ succ_remove + ", C=" + contains + ", SC=" + succ_contains + ", Aborts=" + aborts + ", False Ivals=" + falseinvals;
	}

	@Override
	public void finish() {
		// TODO Auto-generated method stub
		final_set_value = m_set.finish();
	}
}
