//package org.deuce.benchmark.intset;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Random;
import java.util.Stack;
import java.util.TreeMap;
import java.util.concurrent.locks.Lock;

public class SetThread extends BenchmarkThread {
	final int m_threadId;
	final private IntSet m_set;
	final private int ADD = 0;
	final private int REMOVE = 1;
	final private int CONTAINS = 2;

	final int m_range;
	final int m_rate;
	final int m_ops;
	final int m_num_threads;

	int m_nb_add;
	int m_nb_remove;
	int m_nb_succ_add;
	int m_nb_succ_remove;
	int m_nb_succ_contains;
	int m_nb_contains;
	int m_nb_aborts;	
	int m_nb_falseinvals;
	
	int m_thread_contribution = 0;
	
	volatile int delayCount = 0;

	int[] operationType = new int[SetBenchmark.MAX_OPERATIONS_PER_TRANSACTION];
	int[] items = new int[SetBenchmark.MAX_OPERATIONS_PER_TRANSACTION];

	final private Random m_random;

	// //////////////////////////////////////////////////////////////
	// for boosting

	Stack<PTBLinkedList.LogEntry> list_log = new Stack<PTBLinkedList.LogEntry>();
	Stack<PTBSkipList.LogEntry> skiplist_log = new Stack<PTBSkipList.LogEntry>();
	Stack<PTBTree.LogEntry> tree_log = new Stack<PTBTree.LogEntry>();
	HashMap<Integer, Lock> lockSet = new HashMap<Integer, Lock>();
	// for boosting
	// //////////////////////////////////////////////////////////////

	// //////////////////////////////////////////////////////////////
	// for optimistic boosting

	TreeMap<Integer, OTBLinkedList.WriteSetEntry> list_writeset = new TreeMap<Integer, OTBLinkedList.WriteSetEntry>(
			new Comparator<Integer>() {
				@Override
				public int compare(Integer o1, Integer o2) {
					return o2 - o1;
				}
			});
	ArrayList<OTBLinkedList.ReadSetEntry> list_readset = new ArrayList<OTBLinkedList.ReadSetEntry>();

	TreeMap<Integer, OTBSkipList.WriteSetEntry> skiplist_writeset = new TreeMap<Integer, OTBSkipList.WriteSetEntry>(
			new Comparator<Integer>() {
				@Override
				public int compare(Integer o1, Integer o2) {
					return o2 - o1;
				}
			});
	ArrayList<OTBSkipList.ReadSetEntry> skiplist_readset = new ArrayList<OTBSkipList.ReadSetEntry>();

	TreeMap<Integer, OTBTree.WriteSetEntry> tree_writeset = new TreeMap<Integer, OTBTree.WriteSetEntry>(
			new Comparator<Integer>() {
				@Override
				public int compare(Integer o1, Integer o2) {
					return o2 - o1;
				}
			});
	ArrayList<OTBTree.ReadSetEntry> tree_readset = new ArrayList<OTBTree.ReadSetEntry>();
	
	TreeMap<Integer, OTBTreeOptimized.WriteSetEntry> tree_optimized_writeset = new TreeMap<Integer, OTBTreeOptimized.WriteSetEntry>(
			new Comparator<Integer>() {
				@Override
				public int compare(Integer o1, Integer o2) {
					return o2 - o1;
				}
			});
	ArrayList<OTBTreeOptimized.ReadSetEntry> tree_optimized_readset = new ArrayList<OTBTreeOptimized.ReadSetEntry>();

	TreeMap<Integer, OTBTreeOptimizedBackoff.WriteSetEntry> tree_optimizedbackoff_writeset = new TreeMap<Integer, OTBTreeOptimizedBackoff.WriteSetEntry>(
			new Comparator<Integer>() {
				@Override
				public int compare(Integer o1, Integer o2) {
					return o2 - o1;
				}
			});
	ArrayList<OTBTreeOptimizedBackoff.ReadSetEntry> tree_optimizedbackoff_readset = new ArrayList<OTBTreeOptimizedBackoff.ReadSetEntry>();
	
	// for optimistic boosting
	// //////////////////////////////////////////////////////////////

	public SetThread(int id, IntSet set, int range, int rate, int ops, int n) {
		m_num_threads = n;
		m_threadId = id;
		m_set = set;
		m_range = range;
		m_ops = ops;
		m_nb_add = m_nb_remove = m_nb_contains = 0;
		m_rate = rate;
		m_random = new Random();
	}

	protected void step(int phase) {

		for (int c = 0; c < m_ops; c++) {
			int i = m_random.nextInt(100);
			if (i < m_rate / 2)
				operationType[c] = ADD;
			else if (i < m_rate)
				operationType[c] = REMOVE;
			else
				operationType[c] = CONTAINS;

			items[c] = m_random.nextInt(m_range);
//			items[c] = items[c] - items[c]%m_num_threads + m_threadId;
		}

		boolean flag = true;

		while (flag) {
			flag = false;
			try {
				int nb_add = 0, nb_remove = 0, nb_contains = 0, nb_succ_add = 0, nb_succ_remove = 0, nb_succ_contains = 0, contribution = 0;
				for (int c = 0; c < m_ops; c++) {
					if (operationType[c] == ADD) {
						if (m_set.add(items[c]))
						{
							contribution+=items[c];
							if(phase == Benchmark.TEST_PHASE)
								nb_succ_add++;
						}
						if (phase == Benchmark.TEST_PHASE)
							nb_add++;
					} else if (operationType[c] == REMOVE) {
						if (m_set.remove(items[c]))
						{
							contribution-=items[c];
							if(phase == Benchmark.TEST_PHASE)
								nb_succ_remove++;
						}
						if (phase == Benchmark.TEST_PHASE)
							nb_remove++;
					} else {
						if (m_set.contains(items[c])
								&& phase == Benchmark.TEST_PHASE)
							nb_succ_contains++;
						if (phase == Benchmark.TEST_PHASE)
							nb_contains++;
					}
				}
//				if(m_threadId < m_num_threads/2)
//				{
//					for(int c=0;c<1000;c++)
//						delayCount++;
//				}
				
				m_set.commit();
				m_nb_add += nb_add;
				m_nb_remove += nb_remove;
				m_nb_contains += nb_contains;
				m_nb_succ_add += nb_succ_add;
				m_nb_succ_remove += nb_succ_remove;
				m_nb_succ_contains += nb_succ_contains;
				m_thread_contribution += contribution;

			} catch (AbortedException e) {
				flag = true;
				if(phase == Benchmark.TEST_PHASE)
					m_nb_aborts++;
				m_set.abort();
			}
			finally
			{
				// Reset if we are still in the warm-up phase so that we only count the test phase
				if(phase == Benchmark.WARMUP_PHASE)
					m_nb_falseinvals = 0;
			}
		}
	}

	public String getStats() {
		return "A=" + m_nb_add + ", R=" + m_nb_remove + ", SA=" + m_nb_succ_add
				+ ", SR=" + m_nb_succ_remove + ", C=" + m_nb_contains + ", SC="
				+ m_nb_succ_contains + ", Aborts=" + m_nb_aborts + ", False Invals=" + m_nb_falseinvals;
	}
}
